import glglue.frame_input
from OpenGL import GL
from glglue import glo
from glglue.camera.mouse_camera import MouseCamera


class SampleScene:
    def __init__(self) -> None:
        self.initialized = False
        self.mouse_camera = MouseCamera()
        self.drawable = None

    def lazy_initialize(self):
        if self.initialized:
            return
        self.initialized = True

        # shader
        match glo.Shader.load_from_pkg("glglue", "assets/mesh"):
            case str() as error:
                raise Exception(error)
            case glo.Shader() as shader:
                props = shader.create_props(self.mouse_camera.camera)
                # from glglue.drawable import cube
                # self.drawable = cube.create(shader, props)
                from glglue.drawable import teapot

                self.drawable = teapot.create(shader, props)

    def render(self, frame: glglue.frame_input.FrameInput):
        self.lazy_initialize()
        assert self.drawable

        # update camera
        self.mouse_camera.process(frame)

        # https://learnopengl.com/Advanced-OpenGL/Depth-testing
        GL.glEnable(GL.GL_DEPTH_TEST)
        GL.glDepthFunc(GL.GL_LESS)

        # https://learnopengl.com/Advanced-OpenGL/Face-culling
        # GL.glEnable(GL.GL_CULL_FACE)

        # clear
        GL.glViewport(0, 0, frame.width, frame.height)
        r = 0
        g = 0.1 if frame.mouse_left else 0
        if frame.height == 0:
            return
        b = 0
        GL.glClearColor(r, g, b, 1.0)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)  # type: ignore

        # render
        self.drawable.draw()

        # flush
        GL.glFlush()



from typing import NamedTuple, Optional
import math
import logging
import glm
from glglue.camera.mouse_event import DragInterface
from glglue.frame_input import FrameInput

LOGGER = logging.getLogger(__name__)


kEpsilon = 1e-5


class Ray(NamedTuple):
    origin: glm.vec3
    dir: glm.vec3

    def intersect_triangle(
        self, v0: glm.vec3, v1: glm.vec3, v2: glm.vec3
    ) -> Optional[float]:
        """
        https://www.scratchapixel.com/lessons/3d-basic-rendering/ray-tracing-rendering-a-triangle/ray-triangle-intersection-geometric-solution
        """
        # compute plane's normal
        v0v1 = v1 - v0
        v0v2 = v2 - v0
        # no need to normalize
        N = glm.cross(v0v1, v0v2)  # N
        # area2 = N.get_length()

        # Step 1: finding P

        # check if ray and plane are parallel ?
        NdotRayDirection = glm.dot(N, self.dir)
        if math.fabs(NdotRayDirection) < kEpsilon:  # almost 0
            return  # they are parallel so they don't intersect !

        # compute d parameter using equation 2
        d = -glm.dot(N, v0)

        # compute t (equation 3)
        t = -(glm.dot(N, self.origin) + d) / NdotRayDirection
        # check if the triangle is in behind the ray
        if t < 0:
            return  # the triangle is behind

        # compute the intersection point using equation 1
        P = self.origin + self.dir * t

        # Step 2: inside-outside test
        # Vec3f C  # vector perpendicular to triangle's plane

        # edge 0
        edge0 = v1 - v0
        vp0 = P - v0
        C = glm.cross(edge0, vp0)
        if glm.dot(N, C) < 0:
            return  # P is on the right side

        # edge 1
        edge1 = v2 - v1
        vp1 = P - v1
        C = glm.cross(edge1, vp1)
        if glm.dot(N, C) < 0:
            return  # P is on the right side

        # edge 2
        edge2 = v0 - v2
        vp2 = P - v2
        C = glm.cross(edge2, vp2)
        if glm.dot(N, C) < 0:
            return  # P is on the right side

        return t  # this ray hits the triangle


class Perspective:
    def __init__(self, *, near=0.1, far=1000) -> None:
        self.matrix = glm.mat4(1.0)
        self.fov_y = math.pi * 30 / 180
        self.aspect = 1.0
        self.z_near = near
        self.z_far = far
        self.width = 1
        self.height = 1
        self.update_matrix()

    def update_matrix(self) -> None:
        self.matrix = glm.perspectiveRH(
            self.fov_y, self.aspect, self.z_near, self.z_far
        )

    def resize(self, w: int, h: int) -> bool:
        if self.width == w and self.height == h:
            return False
        self.width = w
        self.height = h
        self.aspect = float(w) / h
        self.update_matrix()
        return True


class View:
    def __init__(self, y=0, distance=5) -> None:
        self.gaze = glm.vec3(0, 0, 0)
        self.rotation = glm.quat()
        self.shift = glm.vec3(0, y, -distance)
        self.update_matrix()

    def update_matrix(self):
        g = glm.translate(-self.gaze)
        t = glm.translate(self.shift)
        r = glm.mat4(self.rotation)  # type: ignore
        self.matrix = t * r * g
        self.inverse = glm.inverse(self.matrix)

    def set_gaze(self, gaze: glm.vec3):
        self.gaze = gaze
        # self.matrix = t * r * g
        g = glm.translate(-self.gaze)
        # t = glm.translate(self.shift)
        r = glm.mat4(self.rotation)  # type: ignore
        t = self.matrix * glm.inverse(r * g)
        self.shift = t[3].xyz
        self.update_matrix()


class ScreenShift(DragInterface):
    def __init__(self, view: View, projection: Perspective) -> None:
        self.view = view
        self.projection = projection

    def reset(self, shift: glm.vec3):
        self.view.shift = shift
        self.update()

    def update(self) -> None:
        self.view.update_matrix()

    def begin(self, mouse_input: FrameInput):
        pass

    def drag(self, mouse_input: FrameInput, dx: int, dy: int):
        plane_height = math.tan(self.projection.fov_y * 0.5) * self.view.shift.z * 2
        self.view.shift.x -= dx / self.projection.height * plane_height
        self.view.shift.y += dy / self.projection.height * plane_height
        self.update()

    def end(self, mouse_input: FrameInput):
        pass

    def wheel(self, d: int):
        if d < 0:
            self.view.shift.z *= 1.1
            self.update()
        elif d > 0:
            self.view.shift.z *= 0.9
            self.update()


class TurnTable(DragInterface):
    def __init__(self, view: View) -> None:
        self.view = view
        self.yaw = 0.0
        self.pitch = 0.0
        self.update()

    def update(self) -> None:
        yaw = glm.angleAxis(self.yaw, glm.vec3(0, 1, 0))
        pitch = glm.angleAxis(self.pitch, glm.vec3(1, 0, 0))
        self.view.rotation = pitch * yaw
        self.view.update_matrix()

    def begin(self, mouse_input: FrameInput):
        pass

    def drag(self, mouse_input: FrameInput, dx: int, dy: int):
        self.yaw += dx * 0.01
        self.pitch += dy * 0.01
        self.update()

    def end(self, mouse_input: FrameInput):
        pass


def get_arcball_vector(mouse_input: FrameInput):
    """
    https://en.wikibooks.org/wiki/OpenGL_Programming/Modern_OpenGL_Tutorial_Arcball
    """
    P = glm.vec3(
        mouse_input.mouse_x / mouse_input.width * 2 - 1.0,
        mouse_input.mouse_y / mouse_input.height * 2 - 1.0,
        0,
    )
    P.y = -P.y
    OP_squared = P.x * P.x + P.y * P.y
    if OP_squared <= 1:
        P.z = math.sqrt(1 - OP_squared)  # Pythagoras
    else:
        P = glm.normalize(P)  # nearest point
    return P


class ArcBall(DragInterface):
    def __init__(self, view: View, projection: Perspective) -> None:
        self.view = view
        self.projection = projection
        self.rotation = glm.quat()
        self.tmp_rotation = glm.quat()
        self.x = None
        self.y = None
        self.va = None

    def update(self) -> None:
        self.view.rotation = glm.normalize(self.tmp_rotation * self.rotation)
        self.view.update_matrix()

    def begin(self, mouse_input: FrameInput):
        x = mouse_input.mouse_x
        y = mouse_input.mouse_y
        self.rotation = self.view.rotation
        self.x = x
        self.y = y
        self.va = get_arcball_vector(mouse_input)

    def drag(self, mouse_input: FrameInput, dx: int, dy: int):
        x = mouse_input.mouse_x
        y = mouse_input.mouse_y
        if x == self.x and y == self.y:
            return
        self.x = x
        self.y = y
        vb = get_arcball_vector(mouse_input)
        angle = math.acos(min(1.0, glm.dot(self.va, vb))) * 2  # type: ignore
        axis = glm.cross(self.va, vb)  # type: ignore
        self.tmp_rotation = glm.angleAxis(angle, axis)
        self.update()

    def end(self, mouse_input: FrameInput):
        x = mouse_input.mouse_x
        y = mouse_input.mouse_y
        self.rotation = glm.normalize(self.tmp_rotation * self.rotation)
        self.tmp_rotation = glm.quat()
        self.update()


class Camera:
    def __init__(self, *, near=0.01, far=1000, distance: float = 5, y: float = 0):
        self.projection = Perspective(near=near, far=far)
        self.view = View(y=y, distance=distance)  # type: ignore

    def get_mouse_ray(self, x: int, y: int) -> Ray:
        return get_mouse_ray(
            x,
            y,
            self.projection.width,
            self.projection.height,
            self.view.inverse,
            self.projection.fov_y,
            self.projection.aspect,
        )


def get_mouse_ray(
    x: int, y: int, w: int, h: int, view_inverse: glm.mat4, fov_y: float, aspect: float
) -> Ray:
    origin = view_inverse[3].xyz
    half_fov = fov_y / 2
    dir = view_inverse * glm.vec4(
        (x / w * 2 - 1) * math.tan(half_fov) * (aspect),
        -(y / h * 2 - 1) * math.tan(half_fov),
        -1,
        0,
    )
    return Ray(origin, glm.normalize(dir.xyz))



from .camera import Camera, ArcBall, ScreenShift
from .mouse_event import MouseEvent
from glglue.frame_input import FrameInput


class MouseCamera:
    def __init__(self, *, mouse_event=None, distance: float = 5, y: float = 0) -> None:
        self.camera = Camera(distance=distance, y=y)
        self.mouse_event = mouse_event if mouse_event else MouseEvent()
        self.mouse_event.bind_right_drag(
            ArcBall(self.camera.view, self.camera.projection)
        )
        self.middle_drag = ScreenShift(self.camera.view, self.camera.projection)
        self.mouse_event.bind_middle_drag(self.middle_drag)
        self.mouse_event.wheel += [self.middle_drag.wheel]

    def process(self, frame: FrameInput):
        self.camera.projection.resize(frame.width, frame.height)
        self.mouse_event.process(frame)



# -*- coding: utf-8 -*-
from typing import Callable
import logging
from PySide6 import QtCore, QtGui, QtOpenGLWidgets, QtWidgets
import glglue.frame_input


class Widget(QtOpenGLWidgets.QOpenGLWidget):
    """
    https://doc.qt.io/qtforpython/PySide6/QtOpenGLWidgets/QOpenGLWidget.html
    """

    def __init__(
        self,
        parent,
        render_gl: glglue.frame_input.RenderFunc,
        core_profile=True,
    ):
        super().__init__(parent)
        self.setMouseTracking(True)
        self.render_gl = render_gl
        if core_profile:
            format = QtGui.QSurfaceFormat()
            format.setDepthBufferSize(24)
            format.setStencilBufferSize(8)
            format.setVersion(3, 2)
            format.setProfile(QtGui.QSurfaceFormat.CoreProfile)
            # self.setFormat(format)
            QtGui.QSurfaceFormat.setDefaultFormat(format)
            # must be called before the widget or its parent window gets shown
        self.render_width = 0
        self.render_height = 0
        self.mouse_x = 0
        self.mouse_y = 0
        self.left_down = False
        self.middle_down = False
        self.right_down = False
        self.render_wheel = 0

    def minimumSizeHint(self):
        return QtCore.QSize(50, 50)

    def sizeHint(self):
        return QtCore.QSize(150, 150)

    def paintGL(self):
        self.render_gl(
            glglue.frame_input.FrameInput(
                mouse_x=self.mouse_x,
                mouse_y=self.mouse_y,
                width=self.render_width,
                height=self.render_height,
                mouse_left=self.left_down,
                mouse_middle=self.middle_down,
                mouse_right=self.right_down,
                mouse_wheel=self.render_wheel,
            )
        )
        self.render_wheel = 0

    def resizeGL(self, width, height):
        ratio = QtGui.QGuiApplication.primaryScreen().devicePixelRatio()
        self.render_width = int(width * ratio)
        self.render_height = int(height * ratio)
        self.repaint()

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.left_down = True
            self.repaint()
        elif event.button() == QtCore.Qt.MiddleButton:
            self.middle_down = True
            self.repaint()
        elif event.button() == QtCore.Qt.RightButton:
            self.right_down = True
            self.repaint()

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.left_down = False
            self.repaint()
        elif event.button() == QtCore.Qt.MiddleButton:
            self.middle_down = False
            self.repaint()
        elif event.button() == QtCore.Qt.RightButton:
            self.right_down = False
            self.repaint()

    def mouseMoveEvent(self, event):
        self.mouse_x = event.x()
        self.mouse_y = event.y()
        self.repaint()

    def wheelEvent(self, event):
        self.wheel = event.angleDelta().y()


class CustomLogger(logging.Handler):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def emit(self, record):
        msg = self.format(record)

        if record.levelno == logging.DEBUG:
            msg = f'<font color="gray">{msg}</font><br>'
        elif record.levelno == logging.WARNING:
            msg = f'<font color="orange">{msg}</font><br>'
        elif record.levelno == logging.ERROR:
            msg = f'<font color="red">{msg}</font><br>'
        else:
            msg = f"{msg}<br>"

        self.widget.textCursor().movePosition(QtGui.QTextCursor.Start)
        self.widget.textCursor().insertHtml(msg)

    def write(self, m):
        pass


class QPlainTextEditLogger(QtWidgets.QPlainTextEdit):
    def __init__(self, parent):
        super().__init__(parent)
        self.setReadOnly(True)
        self.log_handler = CustomLogger(self)



import sys
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class Window(Gtk.ApplicationWindow):
    def __init__(self, app):
        super().__init__(application=app)

        from glglue.scene.sample import SampleScene

        self.scene = SampleScene()

        import glglue.gtk3

        self.glWidget = glglue.gtk3.GLArea(self.scene.render)
        self.add(self.glWidget)


def on_activate(app: Gtk.Application):
    window = Window(app)
    window.show_all()


def main():
    app = Gtk.Application()
    app.connect("activate", on_activate)
    app.run()


if __name__ == "__main__":
    main()



from typing import List, Iterable, Optional
from OpenGL import GL
from .vbo import Vbo, Ibo
from .vertex_layout import VertexLayout
import ctypes


class Vao:
    def __init__(
        self, vbo: Vbo, layouts: Iterable[VertexLayout], ibo: Optional[Ibo] = None
    ) -> None:
        self.vao = GL.glGenVertexArrays(1)
        self.vbo = vbo
        self.bind()
        vbo.bind()
        for layout in layouts:
            GL.glEnableVertexAttribArray(layout.attribute.location)
            GL.glVertexAttribPointer(
                layout.attribute.location,
                layout.item_count,
                GL.GL_FLOAT,
                GL.GL_FALSE,
                layout.stride,
                ctypes.c_void_p(layout.byte_offset),
            )
        self.ibo = None
        if ibo:
            self.ibo = ibo
            ibo.bind()
        self.unbind()

    def __del__(self) -> None:
        GL.glDeleteVertexArrays(1, [self.vao])

    def bind(self):
        GL.glBindVertexArray(self.vao)

    def unbind(self):
        GL.glBindVertexArray(0)

    def draw(self, count: int, offset: int = 0, *, topology=GL.GL_TRIANGLES):
        self.bind()
        if self.ibo:
            GL.glDrawElements(topology, count, self.ibo.format, ctypes.c_void_p(offset))
        else:
            GL.glDrawArrays(topology, offset, count)
        self.unbind()



# coding: utf-8
"""
pip install glfw
"""


def main():
    from glglue.scene.sample import SampleScene
    import glglue.glfw

    scene = SampleScene()

    loop = glglue.glfw.LoopManager(title="glfw sample")
    while True:
        frame = loop.begin_frame()
        if not frame:
            break
        scene.render(frame)
        loop.end_frame()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# coding: utf-8
""" [NAME] glglue.glut

[DESCRIPTION] use mainloop(Controller)
"""

from OpenGL.GLUT import *
import time
from logging import getLogger
import sys
import datetime
import glglue.frame_input

LOGGER = getLogger(__name__)


class GlutWindow:
    def __init__(
        self, width: int = 640, height: int = 480, title: bytes = b"glut sample"
    ) -> None:
        """[FUNCTIONS] setup and start glut mainloop"""
        self.mouse_left = False
        self.mouse_right = False
        self.mouse_middle = False
        self.width = 0
        self.height = 0
        self.x = 0
        self.y = 0
        self.wheel = 0

        glutInit(sys.argv)
        glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL)  # type: ignore
        glutInitWindowSize(width, height)
        glutCreateWindow(title)
        # Windowのサイズが変わった時に呼ばれる関数を登録
        glutReshapeFunc(self.resize)
        # 描画時に呼ばれる関数を登録
        glutDisplayFunc(self.draw)
        # マウスボタン押し上げ時に呼ばれる関数
        glutMouseFunc(self.mouse)
        # マウスドラッグ時に呼ばれる関数
        glutMotionFunc(self.motion)
        glutPassiveMotionFunc(self.motion)
        # キーボードが押された時に呼ばれる関数
        glutKeyboardFunc(self.keyboard)

    def draw(self):
        pass

    def resize(self, w, h):
        self.width = w
        self.height = h

    def mouse(self, button, state, x, y):
        if button == GLUT_LEFT_BUTTON:
            if state == GLUT_DOWN:
                self.mouse_left = True
            else:
                self.mouse_left = False
        elif button == GLUT_MIDDLE_BUTTON:
            if state == GLUT_DOWN:
                self.mouse_middle = True
            else:
                self.mouse_middle = False
        elif button == GLUT_RIGHT_BUTTON:
            if state == GLUT_DOWN:
                self.mouse_right = True
            else:
                self.mouse_right = False
        elif button == 3:
            self.wheel = 1
        elif button == 4:
            self.wheel = -1
        else:
            LOGGER.warn("unknown mouse:", button, state, x, y)

        glutPostRedisplay()

    def motion(self, x, y):
        self.x = x
        self.y = y
        glutPostRedisplay()

    def keyboard(self, key, x, y):
        pass


class LoopManager:
    def __init__(self, **kw):
        self.w = GlutWindow(**kw)

    def begin_frame(self) -> glglue.frame_input.FrameInput:
        glutMainLoopEvent()
        clock = time.perf_counter()
        frame = glglue.frame_input.FrameInput(
            elapsed_time=datetime.timedelta(seconds=clock),
            mouse_x=self.w.x,
            mouse_y=self.w.y,
            width=self.w.width,
            height=self.w.height,
            mouse_left=self.w.mouse_left,
            mouse_middle=self.w.mouse_middle,
            mouse_right=self.w.mouse_right,
            mouse_wheel=self.w.wheel,
        )
        self.w.wheel = 0
        return frame

    def end_frame(self):
        glutSwapBuffers()



from typing import Tuple, NamedTuple, List
import glm
from glglue import glo
from .vertex_buffer import MeshBuilder
from .drawable import Drawable

"""
OpenGL default is ccw

+->
|
v
  4 5
  +-+
 / /
+-+
7 6
  0 1
  +-+
 / /
+-+
3 2
"""
S = 0.6
VERTICES = [
    glm.vec3(-S, -S, -S),
    glm.vec3(S, -S, -S),
    glm.vec3(S, -S, S),
    glm.vec3(-S, -S, S),
    glm.vec3(-S, S, -S),
    glm.vec3(S, S, -S),
    glm.vec3(S, S, S),
    glm.vec3(-S, S, S),
]


class Face(NamedTuple):
    indices: Tuple[int, int, int, int]
    color: Tuple[float, float, float]


# CCW
QUADS = [
    Face((7, 3, 2, 6), (0.5, 0.5, 1)),  # front
    Face((5, 1, 0, 4), (0.5, 0.5, 1)),  # back
    Face((6, 2, 1, 5), (1, 0.5, 0.5)),  # right
    Face((4, 0, 3, 7), (1, 0.5, 0.5)),  # left
    Face((4, 7, 6, 5), (0.5, 1, 0.5)),  # top
    Face((0, 1, 2, 3), (0.5, 1, 0.5)),  # bottom
]


def create(shader: glo.Shader, props: List[glo.UniformUpdater]) -> Drawable:
    builder = MeshBuilder()
    for (i0, i1, i2, i3), rgb in QUADS:
        builder.push_quad(
            VERTICES[i0], VERTICES[i1], VERTICES[i2], VERTICES[i3], glm.vec3(*rgb)
        )
    vertices = builder.create_vertices()

    vbo = glo.Vbo()
    vbo.set_vertices(vertices)

    vao = glo.Vao(vbo, glo.VertexLayout.create_list(shader.program))

    drawable = Drawable(vao)
    drawable.push_submesh(shader, len(vertices), props)
    return drawable



from .camera import Camera, ArcBall, ScreenShift, TurnTable, Ray, DragInterface
from .mouse_camera import MouseCamera, MouseEvent

__all__ = [
    "Camera",
    "ArcBall",
    "ScreenShift",
    "TurnTable",
    "Ray",
    "DragInterface",
    "MouseCamera",
    "MouseEvent",
]



# coding: utf-8
"""
Win32APIでOpenGLホストするサンプル。

* Windows専用
* 追加のインストールは不要
"""
import logging

LOGGER = logging.getLogger(__name__)


def main():
    logging.basicConfig(
        format="%(levelname)s:%(name)s:%(message)s", level=logging.DEBUG
    )

    from glglue.scene.sample import SampleScene
    import glglue.wgl

    scene = SampleScene()

    loop = glglue.wgl.LoopManager(width=640, height=480, title=b"sample")

    while True:
        frame = loop.begin_frame()
        if not frame:
            break
        scene.render(frame)
        loop.end_frame()


if __name__ == "__main__":
    main()



import gi
import traceback

gi.require_version("Gtk", "4.0")
from gi.repository import Gtk


class Window(Gtk.ApplicationWindow):
    def __init__(self, app):
        super().__init__(application=app)  # type: ignore

        from glglue.scene.sample import SampleScene
        self.scene = SampleScene()

        import glglue.gtk4

        self.glWidget = glglue.gtk4.GLArea(self.scene.render)
        self.set_child(self.glWidget)


def on_activate(app: Gtk.Application):
    try:
        window = Window(app)
        window.present()
    except Exception as ex:
        traceback.print_exception(ex)
        app.quit()


def main():
    app = Gtk.Application()
    app.connect("activate", on_activate)
    app.run()  # type: ignore


if __name__ == "__main__":
    main()



import ctypes
import struct
import pkgutil
from glglue import glo
from .vertex_buffer import Float3, Vertex
from .drawable import Drawable

WHITE = Float3(1, 1, 1)


class StlTriangle(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("normal", Float3),
        ("position0", Float3),
        ("position1", Float3),
        ("position2", Float3),
        ("attributes", ctypes.c_ushort),
    ]


assert ctypes.sizeof(StlTriangle) == 50


class BinaryReader:
    def __init__(self, data: bytes) -> None:
        self.data = data
        self.pos = 0

    def read(self, size: int) -> bytes:
        if self.pos + size > len(self.data):
            raise RuntimeError()
        data = self.data[self.pos : self.pos + size]
        self.pos += size
        return data

    def read_uint32(self) -> int:
        data = self.read(4)
        return struct.unpack("I", data)[0]

    def read_float32(self) -> float:
        data = self.read(4)
        return struct.unpack("f", data)[0]


def load_teapot() -> ctypes.Array:
    # https://en.wikipedia.org/wiki/Utah_teapot#/media/File:Utah_teapot_(solid).stl
    data = pkgutil.get_data("glglue", "assets/Utah_teapot_(solid).stl")
    assert data

    # https://en.wikipedia.org/wiki/STL_%28file_format%29
    r = BinaryReader(data)

    # UINT8[80]    – Header                 -     80 bytes
    # UINT32       – Number of triangles    -      4 bytes
    header = r.read(80)
    triangle_count = r.read_uint32()

    triangles_type = StlTriangle * triangle_count
    data = r.read(ctypes.sizeof(triangles_type))
    triangles = triangles_type.from_buffer_copy(data)

    vertices = (Vertex * (triangle_count * 3))()

    def push_triangle(i: int, t: StlTriangle, scale: float):
        p0x, p0y, p0z = t.position0
        p1x, p1y, p1z = t.position1
        p2x, p2y, p2z = t.position2
        nx, ny, nz = t.normal

        v0 = vertices[i]
        v0.position = Float3(p0x * scale, p0z * scale, -p0y * scale)
        v0.normal = Float3(nx, nz, -ny)
        v0.color = WHITE

        v1 = vertices[i + 1]
        v1.position = Float3(p1x * scale, p1z * scale, -p1y * scale)
        v1.normal = Float3(nx, nz, -ny)
        v1.color = WHITE

        v2 = vertices[i + 2]
        v2.position = Float3(p2x * scale, p2z * scale, -p2y * scale)
        v2.normal = Float3(nx, nz, -ny)
        v2.color = WHITE

    for i, t in enumerate(triangles):
        push_triangle(i * 3, t, 0.1)

    return vertices


def create(shader, props):
    vertices = load_teapot()

    vbo = glo.Vbo()
    vbo.set_vertices(vertices)

    vao = glo.Vao(vbo, glo.VertexLayout.create_list(shader.program))

    drawable = Drawable(vao)
    drawable.push_submesh(shader, len(vertices), props)
    return drawable



from typing import NamedTuple, Optional, Union, Tuple, cast, List, Callable, TypeAlias
from OpenGL import GL
import glm
import logging

LOGGER = logging.getLogger(__name__)


def GetGLErrorStr(err):
    match (err):
        case GL.GL_NO_ERROR:
            return "No error"
        case GL.GL_INVALID_ENUM:
            return "Invalid enum"
        case GL.GL_INVALID_VALUE:
            return "Invalid value"
        case GL.GL_INVALID_OPERATION:
            return "Invalid operation"
        case GL.GL_STACK_OVERFLOW:
            return "Stack overflow"
        case GL.GL_STACK_UNDERFLOW:
            return "Stack underflow"
        case GL.GL_OUT_OF_MEMORY:
            return "Out of memory"
        case _:
            return "Unknown error"


def CheckGLError():
    while True:
        err = GL.glGetError()
        if GL.GL_NO_ERROR == err:
            break
        LOGGER.error(f"GL Error: {GetGLErrorStr(err)}")


class ShaderCompile:
    def __init__(self, shader_type):
        """
        GL.GL_VERTEX_SHADER
        GL.GL_FRAGMENT_SHADER
        """
        self.shader = GL.glCreateShader(shader_type)

    def compile(self, src: Union[str, bytes]) -> Tuple[bool, str]:
        GL.glShaderSource(self.shader, src, None)
        GL.glCompileShader(self.shader)
        result = GL.glGetShaderiv(self.shader, GL.GL_COMPILE_STATUS)
        if result == GL.GL_TRUE:
            return True, ""
        # error message
        info = GL.glGetShaderInfoLog(self.shader)
        return False, info.decode("ascii")

    def __del__(self):
        GL.glDeleteShader(self.shader)


UniformUpdater: TypeAlias = Callable[[], None]


class Shader:
    def __init__(self) -> None:
        self.program = cast(int, GL.glCreateProgram())

    def __del__(self):
        GL.glDeleteProgram(self.program)

    def __enter__(self):
        self.use()

    def __exit__(self, exc_type, exc_value, traceback):
        self.unuse()
        if exc_type:
            LOGGER.warning(f"{exc_type}: {exc_value}: {traceback}")

    def link(self, vs, fs) -> Tuple[bool, str]:
        GL.glAttachShader(self.program, vs)
        GL.glAttachShader(self.program, fs)
        GL.glLinkProgram(self.program)
        error = GL.glGetProgramiv(self.program, GL.GL_LINK_STATUS)
        if error == GL.GL_TRUE:
            return True, ""

        # error message
        info = GL.glGetProgramInfoLog(self.program)
        return False, info.decode("ascii")

    @staticmethod
    def load(
        vs_src: Union[str, bytes], fs_src: Union[str, bytes]
    ) -> Union["Shader", str]:
        vs = ShaderCompile(GL.GL_VERTEX_SHADER)
        success, info = vs.compile(vs_src)
        if not success:
            return "vs: " + info
        fs = ShaderCompile(GL.GL_FRAGMENT_SHADER)
        success, info = fs.compile(fs_src)
        if not success:
            return "fs: " + info
        shader = Shader()
        success, info = shader.link(vs.shader, fs.shader)
        if not success:
            return "link: " + info
        return shader

    @staticmethod
    def load_from_pkg(pkg: str, name: str) -> Optional["Shader"]:
        import pkgutil

        vs = pkgutil.get_data(pkg, f"{name}.vs")
        assert vs
        fs = pkgutil.get_data(pkg, f"{name}.fs")
        assert fs
        match Shader.load(vs, fs):
            case Shader() as shader:
                return shader
            case str() as info:
                LOGGER.error(f"{pkg}#{name}: {info}")
            case _:
                raise RuntimeError()

    def create_props(self, camera, node=None) -> List[UniformUpdater]:
        from .. import glo

        props: List[Callable[[], None]] = []

        model = UniformLocation.create(self.program, "uModel")
        if model:
            if node:

                def update_model():
                    model.set_mat4(glm.value_ptr(node.world_matrix))

            else:
                identity = glm.mat4(1)

                def update_model():
                    model.set_mat4(glm.value_ptr(identity))

            props.append(update_model)

        view = UniformLocation.create(self.program, "uView")
        if view:

            def update_view():
                view.set_mat4(glm.value_ptr(camera.view.matrix))

            props.append(update_view)

        projection = UniformLocation.create(self.program, "uProjection")
        if projection:

            def update_projection():
                projection.set_mat4(glm.value_ptr(camera.projection.matrix))

            props.append(update_projection)

        return props

    def use(self):
        GL.glUseProgram(self.program)

    def unuse(self):
        GL.glUseProgram(0)


class UniformLocation(NamedTuple):
    name: str
    location: int

    @staticmethod
    def create(program, name: str) -> "UniformLocation":
        location = GL.glGetUniformLocation(program, name)
        if location == -1:
            LOGGER.warn(f"{name}: -1")
        return UniformLocation(name, location)

    def set_int(self, value: int):
        GL.glUniform1i(self.location, value)

    def set_float2(self, value):
        GL.glUniform2fv(self.location, 1, value)

    def set_mat4(self, value, transpose: bool = False, count=1):
        GL.glUniformMatrix4fv(
            self.location, count, GL.GL_TRUE if transpose else GL.GL_FALSE, value
        )


class UniformBlockIndex(NamedTuple):
    name: str
    index: int

    @staticmethod
    def create(program, name: str) -> "UniformBlockIndex":
        index = GL.glGetUniformBlockIndex(program, name)
        return UniformBlockIndex(name, index)



from typing import NamedTuple
from OpenGL import GL


def get_info():
    from OpenGL.GL import (
        glGetString,
        glGetIntegerv,
        GL_VENDOR,
        GL_VERSION,
        GL_SHADING_LANGUAGE_VERSION,
        GL_RENDERER,
        GL_MAJOR_VERSION,
        GL_MINOR_VERSION,
    )

    return {
        "major": glGetIntegerv(GL_MAJOR_VERSION),
        "minor": glGetIntegerv(GL_MINOR_VERSION),
        "vendor": glGetString(GL_VENDOR),
        "version": glGetString(GL_VERSION),
        "shader_language_version": glGetString(GL_SHADING_LANGUAGE_VERSION),
        "renderer": glGetString(GL_RENDERER),
    }


class GLContextHint(NamedTuple):
    major: int = 4
    minor: int = 5
    core_profile: bool = True
    compatible: bool = False


def get_desktop_scaling_factor():
    """
    for high DPI desktop
    """
    import platform

    if platform.platform().startswith("Windows"):
        from ctypes import windll

        desktop = windll.user32.GetDC(0)
        LogicalScreenHeight = windll.gdi32.GetDeviceCaps(desktop, 10)  # VERTRES
        PhysicalScreenHeight = windll.gdi32.GetDeviceCaps(
            desktop, 117
        )  # DESKTOPVERTRES
        return PhysicalScreenHeight / LogicalScreenHeight

    return 1


class DummyScene:
    def __init__(self) -> None:
        pass

    def resize(self, w, h):
        GL.glViewport(0, 0, w, h)

    def draw(self):
        GL.glClearColor(1, 0, 1, 1)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT)



import gi

gi.require_version("Gtk", "4.0")
from gi.repository import Gtk

import glglue.frame_input


class GLArea(Gtk.GLArea):
    def __init__(self, render_callback: glglue.frame_input.RenderFunc) -> None:
        super().__init__()
        self.mouse_x = 0
        self.mouse_y = 0
        self.left_down = False
        self.middle_down = False
        self.right_down = False
        self.wheel = 0

        def on_draw(*args):
            render_callback(
                glglue.frame_input.FrameInput(
                    mouse_x=self.mouse_x,
                    mouse_y=self.mouse_y,
                    width=self.get_width(),
                    height=self.get_height(),
                    mouse_left=self.left_down,
                    mouse_middle=self.middle_down,
                    mouse_right=self.right_down,
                    mouse_wheel=self.wheel,
                )
            )
            self.wheel = 0

        self.connect("render", on_draw)
        self.set_has_depth_buffer(True)

        # motion
        motion_controller = Gtk.EventControllerMotion.new()

        def on_motion(c, x, y):
            self.mouse_x = x
            self.mouse_y = y
            self.queue_render()

        motion_controller.connect("motion", on_motion)
        self.add_controller(motion_controller)

        # left
        left_click_controller = Gtk.GestureClick()
        left_click_controller.set_button(1)

        def on_left_down(*args):
            self.left_down = True
            self.queue_render()

        def on_left_up(*args):
            self.left_down = False
            self.queue_render()

        left_click_controller.connect("pressed", on_left_down)
        left_click_controller.connect("released", on_left_up)
        self.add_controller(left_click_controller)

        # middle
        middle_click_controller = Gtk.GestureClick()
        middle_click_controller.set_button(2)

        def on_middle_down(*args):
            self.middle_down = True
            self.queue_render()

        def on_middle_up(*args):
            self.middle_down = False
            self.queue_render()

        middle_click_controller.connect("pressed", on_middle_down)
        middle_click_controller.connect("released", on_middle_up)
        self.add_controller(middle_click_controller)

        # right
        right_click_controller = Gtk.GestureClick()
        right_click_controller.set_button(3)

        def on_right_down(*args):
            self.right_down = True
            self.queue_render()

        def on_right_up(*args):
            self.right_down = False
            self.queue_render()

        right_click_controller.connect("pressed", on_right_down)
        right_click_controller.connect("released", on_right_up)
        self.add_controller(right_click_controller)

        # wheel
        wheel_controller = Gtk.EventControllerScroll()
        wheel_controller.set_flags(Gtk.EventControllerScrollFlags.VERTICAL)

        def on_wheel(c, x, y):
            self.wheel = -y
            self.queue_render()

        wheel_controller.connect("scroll", on_wheel)
        self.add_controller(wheel_controller)



import argparse


def install_pyside6():
    import pip

    args = ["install", "pyside6"]

    pip.main(args)


def install_pysdl2():
    import pip

    args = ["install", "pysdl2"]
    import platform

    if platform.system() == "Windows":
        args.append("pysdl2-dll")
    pip.main(args)


def install_glfw():
    import pip

    args = ["install", "glfw"]

    pip.main(args)


def main():
    parser = argparse.ArgumentParser(prog="glglue dll installer")

    subparsers = parser.add_subparsers(dest="subparser_name")

    parser_pysdl = subparsers.add_parser("pysdl2")
    parser_glfw = subparsers.add_parser("glfw")
    parser_gyside6 = subparsers.add_parser("pyside6")

    args = parser.parse_args()
    if args.subparser_name == "pysdl2":
        import glglue.pysdl2

        install_pysdl2()
    elif args.subparser_name == "glfw":
        import glglue.glfw

        install_glfw()
    elif args.subparser_name == "pyside6":
        try:
            import glglue.pyside6
        except:
            install_pyside6()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from typing import Optional
from .texture import Texture
from OpenGL import GL
import logging
import ctypes

LOGGER = logging.getLogger(__name__)


class Fbo:
    def __init__(self, width, height, *, use_depth=True) -> None:
        self.texture = Texture(width, height)
        self.fbo = GL.glGenFramebuffers(1)
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fbo)
        GL.glFramebufferTexture2D(
            GL.GL_FRAMEBUFFER,
            GL.GL_COLOR_ATTACHMENT0,
            GL.GL_TEXTURE_2D,
            self.texture.handle,
            0,
        )
        GL.glDrawBuffers([GL.GL_COLOR_ATTACHMENT0])

        if use_depth:
            self.depth = GL.glGenRenderbuffers(1)
            GL.glBindRenderbuffer(GL.GL_RENDERBUFFER, self.depth)
            GL.glRenderbufferStorage(
                GL.GL_RENDERBUFFER, GL.GL_DEPTH_COMPONENT, width, height
            )
            GL.glFramebufferRenderbuffer(
                GL.GL_FRAMEBUFFER,
                GL.GL_DEPTH_ATTACHMENT,
                GL.GL_RENDERBUFFER,
                self.depth,
            )
        else:
            self.depth = 0

        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)
        LOGGER.debug(f"fbo: {self.fbo}, texture: {self.texture}, depth: {self.depth}")

    def __del__(self):
        LOGGER.debug(f"fbo: {self.fbo}")
        GL.glDeleteFramebuffers(1, [self.fbo])

    def bind(self):
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fbo)

    def unbind(self):
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)


class FboRenderer:
    """
    https://qiita.com/edo_m18/items/95483cabf50494f53bb5
    """

    def __init__(self) -> None:
        self.fbo: Optional[Fbo] = None

    def clear(self, width, height, color):
        if width == 0 or height == 0:
            return 0

        if self.fbo:
            if self.fbo.texture.width != width or self.fbo.texture.height != height:
                del self.fbo
                self.fbo = None
        if not self.fbo:
            self.fbo = Fbo(width, height)

        self.fbo.bind()
        GL.glViewport(0, 0, width, height)
        GL.glScissor(0, 0, width, height)
        GL.glClearColor(
            color[0] * color[3], color[1] * color[3], color[2] * color[3], color[3]
        )
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)  # type: ignore
        GL.glClearDepth(1.0)
        GL.glDepthFunc(GL.GL_LESS)
        return ctypes.c_void_p(int(self.fbo.texture.handle))



from OpenGL import GL
import logging
import ctypes

LOGGER = logging.getLogger(__name__)


class Vbo:
    def __init__(self):
        self.vbo = GL.glGenBuffers(1)

    def __del__(self):
        LOGGER.debug(f"delete vbo: {self.vbo}")
        GL.glDeleteBuffers(1, [self.vbo])

    def bind(self):
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)

    def unbind(self):
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)

    def set_vertices(self, vertices, *, is_dynamic: bool = False):
        self.bind()
        GL.glBufferData(
            GL.GL_ARRAY_BUFFER,
            ctypes.sizeof(vertices),
            vertices,
            GL.GL_DYNAMIC_DRAW if is_dynamic else GL.GL_STATIC_DRAW,
        )
        self.unbind()

    def update(self, vertices, offset=0) -> None:
        self.bind()
        GL.glBufferSubData(
            GL.GL_ARRAY_BUFFER, offset, ctypes.sizeof(vertices), vertices
        )
        self.unbind()


class Ibo:
    def __init__(self):
        self.vbo = GL.glGenBuffers(1)
        self.format = 0

    def __del__(self):
        LOGGER.debug(f"delete vbo: {self.vbo}")
        GL.glDeleteBuffers(1, [self.vbo])

    def bind(self):
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.vbo)

    def unbind(self):
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0)

    def set_indices(self, indices: ctypes.Array, *, is_dynamic: bool = False):
        match indices._type_:
            case ctypes.c_ushort:
                self.format = GL.GL_UNSIGNED_SHORT
            case ctypes.c_uint:
                self.format = GL.GL_UNSIGNED_INT
            case _:
                raise NotImplementedError()
        self.bind()
        GL.glBufferData(
            GL.GL_ELEMENT_ARRAY_BUFFER,
            ctypes.sizeof(indices),
            indices,
            GL.GL_DYNAMIC_DRAW if is_dynamic else GL.GL_STATIC_DRAW,
        )
        self.unbind()

    def update(self, indices, offset=0) -> None:
        self.bind()
        GL.glBufferSubData(
            GL.GL_ELEMENT_ARRAY_BUFFER, offset, ctypes.sizeof(indices), indices
        )
        self.unbind()



import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk, Gdk

import glglue.frame_input


class GLArea(Gtk.GLArea):
    def __init__(self, render_callback: glglue.frame_input.RenderFunc) -> None:
        super().__init__()

        self.mouse_x = 0
        self.mouse_y = 0
        self.left_down = False
        self.middle_down = False
        self.right_down = False
        self.wheel = 0

        def on_draw(*args):
            render_callback(
                glglue.frame_input.FrameInput(
                    mouse_x=self.mouse_x,
                    mouse_y=self.mouse_y,
                    width=self.get_allocated_width(),
                    height=self.get_allocated_height(),
                    mouse_left=self.left_down,
                    mouse_middle=self.middle_down,
                    mouse_right=self.right_down,
                    mouse_wheel=self.wheel,
                )
            )
            self.wheel = 0

        self.connect("render", on_draw)
        self.set_has_depth_buffer(True)
        self.set_double_buffered(False)

        # motion
        self.add_events(
            Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
            | Gdk.EventMask.SCROLL_MASK
        )

        def on_motion(w, e: Gdk.EventButton):
            self.mouse_x = e.x
            self.mouse_y = e.y
            self.queue_render()

        self.connect("motion-notify-event", on_motion)

        # button press
        def on_press(w, e: Gdk.EventButton):
            match e.button:
                case 1:
                    self.left_down = True
                case 2:
                    self.middle_down = True
                case 3:
                    self.right_down = True
            self.queue_render()

        self.connect("button-press-event", on_press)

        # button release
        def on_release(w, e: Gdk.EventButton):
            match e.button:
                case 1:
                    self.left_down = False
                case 2:
                    self.middle_down = False
                case 3:
                    self.right_down = False
            self.queue_render()

        self.connect("button-release-event", on_release)

        # wheel
        def on_wheel(w, e: Gdk.EventScroll):
            match e.direction:
                case Gdk.ScrollDirection.UP:
                    self.wheel = 1
                case Gdk.ScrollDirection.DOWN:
                    self.wheel = -1
            self.queue_render()

        self.connect("scroll-event", on_wheel)



# coding: utf-8
"""
pip install pyside6
"""
from PySide6 import QtWidgets


class Window(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__(None)
        import glglue.pyside6
        from glglue.scene.sample import SampleScene

        self.scene = SampleScene()

        self.glwidget = glglue.pyside6.Widget(self, render_gl=self.scene.render)
        self.setCentralWidget(self.glwidget)


def main():
    import sys

    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    window.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



from typing import NamedTuple, Iterable, List
from OpenGL import GL
import logging

LOGGER = logging.getLogger(__name__)


class AttributeLocation(NamedTuple):
    name: str
    location: int

    @staticmethod
    def create(program, name: str) -> "AttributeLocation":
        location = GL.glGetAttribLocation(program, name)
        assert location != -1
        return AttributeLocation(name, location)


type_map = {
    GL.GL_FLOAT: (GL.GL_FLOAT, 1),
    GL.GL_FLOAT_VEC2: (GL.GL_FLOAT, 2),
    GL.GL_FLOAT_VEC3: (GL.GL_FLOAT, 3),
    GL.GL_FLOAT_VEC4: (GL.GL_FLOAT, 4),
    GL.GL_FLOAT_MAT2: (GL.GL_FLOAT, 4),
    GL.GL_FLOAT_MAT3: (GL.GL_FLOAT, 9),
    GL.GL_FLOAT_MAT4: (GL.GL_FLOAT, 16),
    GL.GL_FLOAT_MAT2x3: (GL.GL_FLOAT, 6),
    GL.GL_FLOAT_MAT2x4: (GL.GL_FLOAT, 8),
    GL.GL_FLOAT_MAT3x2: (GL.GL_FLOAT, 6),
    GL.GL_FLOAT_MAT3x4: (GL.GL_FLOAT, 12),
    GL.GL_FLOAT_MAT4x2: (GL.GL_FLOAT, 8),
    GL.GL_FLOAT_MAT4x3: (GL.GL_FLOAT, 12),
    GL.GL_INT: (GL.GL_INT, 1),
    GL.GL_INT_VEC2: (GL.GL_INT, 2),
    GL.GL_INT_VEC3: (GL.GL_INT, 3),
    GL.GL_INT_VEC4: (GL.GL_INT, 4),
    GL.GL_UNSIGNED_INT: (GL.GL_UNSIGNED_INT, 1),
    GL.GL_UNSIGNED_INT_VEC2: (GL.GL_UNSIGNED_INT, 2),
    GL.GL_UNSIGNED_INT_VEC3: (GL.GL_UNSIGNED_INT, 3),
    GL.GL_UNSIGNED_INT_VEC4: (GL.GL_UNSIGNED_INT, 4),
    GL.GL_DOUBLE: (GL.GL_DOUBLE, 1),
    GL.GL_DOUBLE_VEC2: (GL.GL_DOUBLE, 2),
    GL.GL_DOUBLE_VEC3: (GL.GL_DOUBLE, 3),
    GL.GL_DOUBLE_VEC4: (GL.GL_DOUBLE, 4),
    GL.GL_DOUBLE_MAT2: (GL.GL_DOUBLE, 4),
    GL.GL_DOUBLE_MAT3: (GL.GL_DOUBLE, 9),
    GL.GL_DOUBLE_MAT4: (GL.GL_DOUBLE, 16),
    GL.GL_DOUBLE_MAT2x3: (GL.GL_DOUBLE, 6),
    GL.GL_DOUBLE_MAT2x4: (GL.GL_DOUBLE, 8),
    GL.GL_DOUBLE_MAT3x2: (GL.GL_DOUBLE, 6),
    GL.GL_DOUBLE_MAT3x4: (GL.GL_DOUBLE, 12),
    GL.GL_DOUBLE_MAT4x2: (GL.GL_DOUBLE, 8),
    GL.GL_DOUBLE_MAT4x3: (GL.GL_DOUBLE, 12),
}


def byte_size(t, n):
    match t:
        case GL.GL_FLOAT:
            return 4 * n
        case _:
            raise NotImplemented()


def to_str(src) -> str:
    match src:
        case str():
            return src
        case bytes():
            return src.decode("ascii")
        case _:
            data = bytes(src)
            pos = data.index(0)
            if pos != -1:
                data = data[:pos]
            return data.decode("ascii")


class VertexLayout(NamedTuple):
    attribute: AttributeLocation
    item_count: int  # maybe float1, 2, 3, 4 and 16
    stride: int
    byte_offset: int

    @staticmethod
    def create_list(program) -> List["VertexLayout"]:
        count = GL.glGetProgramiv(program, GL.GL_ACTIVE_ATTRIBUTES)
        LOGGER.debug(f"Active Attributes: {count}")

        layouts: List[VertexLayout] = []
        stride = 0
        for i in range(count):
            name, size, type_ = GL.glGetActiveAttrib(program, i)
            # https://www.khronos.org/registry/OpenGL-Refpages/gl4/html/glGetActiveAttrib.xhtml
            attribute = AttributeLocation.create(program, to_str(name))
            match type_map.get(type_):
                case (t, n):
                    size = byte_size(t, n)
                    layouts.append(VertexLayout(attribute, n, 0, size))
                    stride += size
                case _:
                    raise RuntimeError()
        # not same with location order
        layouts.sort(key=lambda l: l.attribute.location)

        result = []
        offset = 0
        for layout in layouts:
            result.append(
                VertexLayout(layout.attribute, layout.item_count, stride, offset)
            )
            offset += layout.byte_offset
        return result



# coding: utf-8
"""
https://github.com/mcfletch/pyopengl/tree/master/OpenGL/DLLS

environment variable PATH to freeglut64.vc14.dll
"""


def main():

    import glglue.glut
    from glglue.scene.sample import SampleScene

    scene = SampleScene()

    # manual loop
    loop = glglue.glut.LoopManager()
    while True:
        # require freeglut glutMainLoopEvent
        frame = loop.begin_frame()
        if not frame:
            break
        scene.render(frame)
        loop.end_frame()


if __name__ == "__main__":
    main()



from typing import Optional
import sys
from logging import getLogger
from OpenGL import GL
import OpenGL.error
from ctypes import wintypes
import ctypes
import glglue.frame_input
import datetime

from . import win32con

logger = getLogger(__name__)


##############################################################################
# Windows types
##############################################################################
def ErrorIfZero(handle):
    if handle == 0:
        raise ctypes.WinError()
    else:
        return handle


def LOWORD(n):
    return n & 0xFFFF


def HIWORD(n):
    return (n >> 16) & 0xFFFF


WNDPROC = ctypes.WINFUNCTYPE(
    ctypes.c_int, wintypes.HWND, ctypes.c_uint, wintypes.WPARAM, wintypes.LPARAM
)


class WNDCLASS(ctypes.Structure):
    _fields_ = [
        ("style", ctypes.c_uint),
        ("lpfnWndProc", WNDPROC),
        ("cbClsExtra", ctypes.c_int),
        ("cbWndExtra", ctypes.c_int),
        ("hInstance", wintypes.HANDLE),
        ("hIcon", wintypes.HANDLE),
        ("hCursor", wintypes.HANDLE),
        ("hbrBackground", wintypes.HANDLE),
        ("lpszMenuName", wintypes.LPCSTR),
        ("lpszClassName", wintypes.LPCSTR),
    ]


class CREATESTRUCT(ctypes.Structure):
    _fields_ = [
        ("lpCreateParams", ctypes.c_void_p),
        ("hInstance", wintypes.HANDLE),
        ("hMenu", wintypes.HANDLE),
        ("hwndParent", wintypes.HWND),
        ("cy", ctypes.c_int),
        ("cx", ctypes.c_int),
        ("y", ctypes.c_int),
        ("x", ctypes.c_int),
        ("style", ctypes.c_int),
        ("lpszName", wintypes.LPCSTR),
        ("lpszClass", wintypes.LPCSTR),
        ("dwExStyle", ctypes.c_int),
    ]


class PIXELFORMATDESCRIPTOR(ctypes.Structure):
    _fields_ = [
        ("nSize", ctypes.c_ushort),
        ("nVersion", ctypes.c_ushort),
        ("dwFlags", ctypes.c_ulong),
        ("iPixelType", ctypes.c_ubyte),
        ("cColorBits", ctypes.c_ubyte),
        ("cRedBits", ctypes.c_ubyte),
        ("cRedShift", ctypes.c_ubyte),
        ("cGreenBits", ctypes.c_ubyte),
        ("cGreenShift", ctypes.c_ubyte),
        ("cBlueBits", ctypes.c_ubyte),
        ("cBlueShift", ctypes.c_ubyte),
        ("cAlphaBits", ctypes.c_ubyte),
        ("cAlphaShift", ctypes.c_ubyte),
        ("cAccumBits", ctypes.c_ubyte),
        ("cAccumRedBits", ctypes.c_ubyte),
        ("cAccumGreenBits", ctypes.c_ubyte),
        ("cAccumBlueBits", ctypes.c_ubyte),
        ("cAccumAlphaBits", ctypes.c_ubyte),
        ("cDepthBits", ctypes.c_ubyte),
        ("cStencilBits", ctypes.c_ubyte),
        ("cAuxBuffers", ctypes.c_ubyte),
        ("iLayerType", ctypes.c_ubyte),
        ("bReserved", ctypes.c_ubyte),
        ("dwLayerMask", ctypes.c_ulong),
        ("dwVisibleMask", ctypes.c_ulong),
        ("dwDamageMask", ctypes.c_ulong),
    ]


class RECT(ctypes.Structure):
    _fields_ = [
        ("left", ctypes.c_long),
        ("top", ctypes.c_long),
        ("right", ctypes.c_long),
        ("bottom", ctypes.c_long),
    ]


class PAINTSTRUCT(ctypes.Structure):
    _fields_ = [
        ("hdc", wintypes.HDC),
        ("fErase", wintypes.BOOL),
        ("rcPaint", RECT),
        ("fRestore", wintypes.BOOL),
        ("fIncUpdate", wintypes.BOOL),
        ("rgbReserved", ctypes.c_char * 32),
    ]


class POINT(ctypes.Structure):
    _fields_ = [("x", ctypes.c_long), ("y", ctypes.c_long)]


class MSG(ctypes.Structure):
    _fields_ = [
        ("hwnd", wintypes.HWND),
        ("message", ctypes.c_uint),
        ("wParam", wintypes.WPARAM),
        ("lParam", wintypes.LPARAM),
        ("time", ctypes.c_int),
        ("pt", POINT),
    ]


ShowWindow = ctypes.windll.user32.ShowWindow
ShowWindow.argtypes = [wintypes.HWND, ctypes.c_int]
ShowWindow.restype = wintypes.BOOL

RegisterClass = ctypes.windll.user32.RegisterClassA
RegisterClass.argtypes = [ctypes.c_void_p]
RegisterClass.restype = wintypes.ATOM

CreateWindowEx = ctypes.windll.user32.CreateWindowExA
CreateWindowEx.argtypes = [
    ctypes.c_int,
    wintypes.LPCSTR,
    wintypes.LPCSTR,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    wintypes.HWND,
    wintypes.HANDLE,
    wintypes.HANDLE,
    ctypes.py_object,
]
CreateWindowEx.restype = wintypes.HWND

if sys.maxsize > 2**32:
    # 64bit
    SetWindowLongPtr = ctypes.windll.user32.SetWindowLongPtrA
    SetWindowLongPtr.argtypes = [wintypes.HWND, ctypes.c_int, ctypes.c_void_p]
    SetWindowLongPtr.restype = ctypes.c_void_p

    GetWindowLongPtr = ctypes.windll.user32.GetWindowLongPtrA
    GetWindowLongPtr.argtypes = [wintypes.HWND, ctypes.c_int]
    GetWindowLongPtr.restype = ctypes.c_void_p
else:
    # 64bit
    SetWindowLongPtr = ctypes.windll.user32.SetWindowLongA
    SetWindowLongPtr.argtypes = [wintypes.HWND, ctypes.c_int, ctypes.c_void_p]
    SetWindowLongPtr.restype = ctypes.c_void_p

    GetWindowLongPtr = ctypes.windll.user32.GetWindowLongA
    GetWindowLongPtr.argtypes = [wintypes.HWND, ctypes.c_int]
    GetWindowLongPtr.restype = ctypes.c_void_p

GetDC = ctypes.windll.user32.GetDC
GetDC.argtypes = [wintypes.HWND]

BeginPaint = ctypes.windll.user32.BeginPaint
BeginPaint.argtypes = [wintypes.HWND, ctypes.c_void_p]
BeginPaint.restype = wintypes.HDC

EndPaint = ctypes.windll.user32.EndPaint
EndPaint.argtypes = [wintypes.HWND, ctypes.c_void_p]
EndPaint.restype = wintypes.HDC

ChoosePixelFormat = ctypes.windll.gdi32.ChoosePixelFormat
ChoosePixelFormat.argtypes = [wintypes.HDC, ctypes.c_void_p]
ChoosePixelFormat.restype = ErrorIfZero

SetPixelFormat = ctypes.windll.gdi32.SetPixelFormat
SetPixelFormat.argtypes = [
    wintypes.HDC,
    ctypes.c_int,
    ctypes.c_void_p,
]
SetPixelFormat.restype = ErrorIfZero

SwapBuffers = ctypes.windll.gdi32.SwapBuffers
SwapBuffers.argtypes = [
    wintypes.HDC,
]
SwapBuffers.restype = ErrorIfZero

wglCreateContext = ctypes.windll.opengl32.wglCreateContext
wglCreateContext.argtypes = [
    wintypes.HDC,
]
wglCreateContext.restype = ErrorIfZero

wglDeleteContext = ctypes.windll.opengl32.wglDeleteContext

wglMakeCurrent = ctypes.windll.opengl32.wglMakeCurrent
wglMakeCurrent.argtypes = [
    wintypes.HDC,
    wintypes.HANDLE,
]
wglMakeCurrent.restype = ErrorIfZero

wglGetProcAddress = ctypes.windll.opengl32.wglGetProcAddress
wglGetProcAddress.argtypes = [ctypes.c_void_p]
wglGetProcAddress.restype = ErrorIfZero

DefWindowProc = ctypes.windll.user32.DefWindowProcA
DefWindowProc.argtypes = [wintypes.HWND, ctypes.c_int, wintypes.WPARAM, wintypes.LPARAM]
DefWindowProc.restype = ctypes.c_int

timeGetTime = ctypes.windll.winmm.timeGetTime
timeGetTime.argtypes = []
timeGetTime.restype = ctypes.c_ulong

##############################################################################
# api class
##############################################################################


class Window(object):
    def __init__(self):
        self.hwnd = None
        self.hrc = None
        self.ptrs = {}
        self.callbacks = {
            win32con.WM_PAINT: self.onPaint,
            win32con.WM_SIZE: self.onSize,
            win32con.WM_ERASEBKGND: self.onErasebkgnd,
            win32con.WM_DESTROY: self.onDestroy,
            win32con.WM_LBUTTONDOWN: self.onLeftDown,
            win32con.WM_LBUTTONUP: self.onLeftUp,
            win32con.WM_MBUTTONDOWN: self.onMiddleDown,
            win32con.WM_MBUTTONUP: self.onMiddleUp,
            win32con.WM_RBUTTONDOWN: self.onRightDown,
            win32con.WM_RBUTTONUP: self.onRightUp,
            win32con.WM_MOUSEMOVE: self.onMouseMove,
            win32con.WM_MOUSEWHEEL: self.onMouseWheel,
            win32con.WM_KEYDOWN: self.onKeyDown,
        }
        self.mouse_left = False
        self.mouse_right = False
        self.mouse_middle = False
        self.mouse_capture = False
        self.width = 0
        self.height = 0
        self.x = 0
        self.y = 0
        self.wheel = 0

    def finalize(self):
        logger.debug("finalize %s", self.__class__)

    def update_mouse(self, left, middle, right):
        self.mouse_left = left
        self.mouse_middle = middle
        self.mouse_right = right
        if left or middle or right:
            if not self.mouse_capture:
                ctypes.windll.user32.SetCapture(self.hwnd)
                self.mouse_capture = True
        else:
            if self.mouse_capture:
                ctypes.windll.user32.ReleaseCapture()
                self.mouse_capture = False

    def Redraw(self):
        ctypes.windll.user32.InvalidateRect(self.hwnd, ctypes.c_int(0), ctypes.c_int(0))

    def Show(self):
        ShowWindow(self.hwnd, win32con.SW_SHOWNORMAL)
        ctypes.windll.user32.UpdateWindow(self.hwnd)

    def SetLongPtr(self, hwnd, key, value):
        self.ptrs[key] = value
        SetWindowLongPtr(hwnd, key, value)

    def onPaint(self, hwnd, message, wParam, lParam):
        ps = PAINTSTRUCT()
        hdc = BeginPaint(hwnd, ctypes.byref(ps))
        # if self.hrc and self.controller:
        #     self.controller.draw()
        #     SwapBuffers(hdc)
        EndPaint(hwnd, ctypes.byref(ps))
        return 0

    def onSize(self, hwnd, message, wParam, lParam):
        self.width = LOWORD(lParam)
        self.height = HIWORD(lParam)
        # logger.debug("WM_SIZE %d x %d %s" % (w, h, self.controller))
        return 0

    def onErasebkgnd(self, hwnd, message, wParam, lParam):
        return 0

    def onDestroy(self, hwnd, message, wParam, lParam):
        # sys.stderr.write("WM_DESTROY\n")
        ctypes.windll.user32.PostQuitMessage(0)
        return 0

    def onKeyDown(self, hwnd, message, wParam, lParam):
        # if self.controller:
        #     if wParam >= 65 and wParam <= 90:
        #         # lower
        #         if self.controller.onKeyDown(wParam + 32):
        #             self.Redraw()
        #     else:
        #         if self.controller.onKeyDown(wParam):
        #             self.Redraw()
        return 0

    def onMouseMove(self, hwnd, message, wParam, lParam):
        self.x = LOWORD(lParam)
        self.y = HIWORD(lParam)
        self.Redraw()
        return 0

    def onMouseWheel(self, hwnd, message, wParam, lParam):
        self.wheel = ctypes.c_short(HIWORD(wParam)).value
        self.Redraw()
        return 0

    def onLeftUp(self, hwnd, message, wParam, lParam):
        self.update_mouse(False, self.mouse_middle, self.mouse_right)
        self.Redraw()
        return 0

    def onLeftDown(self, hwnd, message, wParam, lParam):
        self.update_mouse(True, self.mouse_middle, self.mouse_right)
        self.Redraw()
        return 0

    def onMiddleUp(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, False, self.mouse_right)
        self.Redraw()
        return 0

    def onMiddleDown(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, True, self.mouse_right)
        self.Redraw()
        return 0

    def onRightUp(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, self.mouse_middle, False)
        self.Redraw()
        return 0

    def onRightDown(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, self.mouse_middle, True)
        self.Redraw()
        return 0

    def WndProc(self, hwnd, message, wParam, lParam):
        if message in self.callbacks:
            return self.callbacks[message](hwnd, message, wParam, lParam)
        else:
            return DefWindowProc(hwnd, message, wParam, lParam)

    def createGLContext(self, bits, version=None, profile="compatibility"):
        hdc = GetDC(self.hwnd)

        pfd = PIXELFORMATDESCRIPTOR()
        pfd.nsize = ctypes.sizeof(PIXELFORMATDESCRIPTOR)
        pfd.nVersion = 1
        pfd.dwFlags = (
            win32con.PFD_DRAW_TO_WINDOW
            | win32con.PFD_SUPPORT_OPENGL
            | win32con.PFD_DOUBLEBUFFER
        )
        pfd.iPixelType = win32con.PFD_TYPE_RGBA
        pfd.cColorBits = bits
        pfd.cRedBits = 0
        pfd.cRedShift = 0
        pfd.cGreenBits = 0
        pfd.cGreenShift = 0
        pfd.cBlueBits = 0
        pfd.cBlueShift = 0
        pfd.cAlphaBits = 0
        pfd.cAlphaShift = 0
        pfd.cAccumBits = 0
        pfd.cAccumRedBits = 0
        pfd.cAccumGreenBits = 0
        pfd.cAccumBlueBits = 0
        pfd.cAccumAlphaBits = 0
        pfd.cDepthBits = bits
        pfd.cStencilBits = 2
        pfd.cAuxBuffers = 0
        pfd.iLayerType = win32con.PFD_MAIN_PLANE
        pfd.bReserved = 0
        pfd.dwLayerMask = 0
        pfd.dwVisibleMask = 0
        pfd.dwDamageMask = 0

        pixelFormat = ChoosePixelFormat(hdc, ctypes.byref(pfd))
        SetPixelFormat(hdc, pixelFormat, ctypes.byref(pfd))

        self.hrc = wglCreateContext(hdc)
        wglMakeCurrent(hdc, self.hrc)

        if version or profile:
            p = wglGetProcAddress("wglCreateContextAttribsARB".encode("utf-8"))
            if not p:
                raise Exception()
            wglCreateContextAttribsARB = ctypes.WINFUNCTYPE(
                ctypes.c_void_p,
                wintypes.HDC,
                ctypes.c_void_p,
                ctypes.POINTER(ctypes.c_int),
            )(p)

            attribs = []
            if version:
                major, minor = version.split(".")
                attribs.append(win32con.WGL_CONTEXT_MAJOR_VERSION_ARB)
                attribs.append(int(major))
                attribs.append(win32con.WGL_CONTEXT_MINOR_VERSION_ARB)
                attribs.append(int(minor))

            attribs.append(win32con.WGL_CONTEXT_FLAGS_ARB)
            attribs.append(0)

            if profile == "core":
                attribs.append(win32con.WGL_CONTEXT_PROFILE_MASK_ARB)
                attribs.append(win32con.WGL_CONTEXT_CORE_PROFILE_BIT_ARB)
            elif profile == "compatibility":
                attribs.append(win32con.WGL_CONTEXT_PROFILE_MASK_ARB)
                attribs.append(win32con.WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB)

            attribs.append(0)

            array = (ctypes.c_int * len(attribs))(*attribs)
            hrc = wglCreateContextAttribsARB(hdc, None, array)
            wglDeleteContext(self.hrc)
            self.hrc = hrc
            wglMakeCurrent(hdc, self.hrc)

        for x in [
            "GL_VERSION",
            # 'GL_EXTENSIONS',
            "GL_SHADING_LANGUAGE_VERSION",
            "GL_RENDERER",
        ]:
            logger.info("%s: %s", x, GL.glGetString(getattr(GL, x)))

        # GL_ARB_compatibility
        profile = None
        try:
            extensions = (
                GL.glGetString(GL.GL_EXTENSIONS).decode("ascii").split()  # type: ignore
            )
            if "GL_ARB_compatibility" in extensions:
                profile = "COMPATIBILITY"
        except OpenGL.error.GLerror as ex:
            profile = "CORE"
        logger.info(
            "OpenGL%d.%d %s",
            GL.glGetInteger(GL.GL_MAJOR_VERSION),
            GL.glGetInteger(GL.GL_MINOR_VERSION),
            profile,
        )

    @staticmethod
    def WndProcProxy(hwnd, message, wParam, lParam):
        # sys.stderr.write("Window::WndProcProxy\n")
        p = GetWindowLongPtr(hwnd, win32con.GWL_USERDATA)
        window = ctypes.cast(p, ctypes.py_object)
        return window.value.WndProc(hwnd, message, wParam, lParam)


class WindowFactory(object):
    def __init__(self):
        self.classes = []
        self.windows = []
        self.wndclass = self.register_class(b"MainWin")

    def __del__(self):
        self.finalize()

    def finalize(self):
        for w in self.windows:
            w.value.finalize()
        logger.debug("finalize %s", self.__class__)

    def register_class(self, className):
        """
        RegisterClassA
        """
        # Define Window Class
        wndclass = WNDCLASS()
        wndclass.style = win32con.CS_HREDRAW | win32con.CS_VREDRAW
        wndclass.lpfnWndProc = WNDPROC(WindowFactory.WndProc)
        wndclass.cbClsExtra = wndclass.cbWndExtra = 0
        wndclass.hInstance = ctypes.windll.kernel32.GetModuleHandleA(
            wintypes.HANDLE(win32con.NULL)
        )
        wndclass.hIcon = ctypes.windll.user32.LoadIconA(
            wintypes.HANDLE(win32con.NULL), wintypes.LPCSTR(win32con.IDI_APPLICATION)
        )
        wndclass.hCursor = ctypes.windll.user32.LoadCursorA(
            wintypes.HANDLE(win32con.NULL), wintypes.LPCSTR(win32con.IDC_ARROW)
        )
        # wndclass.hbrBackground = ctypes.windll.gdi32.GetStockObject(
        #     ctypes.c_int(win32con.WHITE_BRUSH))
        wndclass.hbrBackground = None
        wndclass.lpszMenuName = None
        wndclass.lpszClassName = className
        # Register Window Class
        res = RegisterClass(ctypes.byref(wndclass))
        # logger.debug(res)
        if not res:
            raise ctypes.WinError()
        self.classes.append(wndclass)
        return wndclass

    def create(
        self,
        klass,
        width=win32con.CW_USEDEFAULT,
        height=win32con.CW_USEDEFAULT,
        wndclass=None,
        title=b"glglue.wgl",
    ):
        if not wndclass:
            wndclass = self.wndclass

        window = klass()
        pywindow = ctypes.py_object(window)
        hwnd = CreateWindowEx(
            0,
            wndclass.lpszClassName,
            title,
            win32con.WS_OVERLAPPEDWINDOW,
            win32con.CW_USEDEFAULT,
            win32con.CW_USEDEFAULT,
            width,
            height,
            win32con.NULL,
            win32con.NULL,
            wndclass.hInstance,
            pywindow,
        )
        self.windows.append(pywindow)
        return window

    def loop(self, window):
        msg = MSG()
        pMsg = ctypes.pointer(msg)
        NULL = ctypes.c_int(win32con.NULL)
        # while ctypes.windll.user32.GetMessageA( pMsg, NULL, 0, 0) != 0:
        #    ctypes.windll.user32.TranslateMessage(pMsg)
        #    ctypes.windll.user32.DispatchMessageA(pMsg)
        lastCount = timeGetTime()
        while True:
            if (
                ctypes.windll.user32.PeekMessageA(
                    pMsg, NULL, 0, 0, win32con.PM_NOREMOVE
                )
                != 0
            ):
                if ctypes.windll.user32.GetMessageA(pMsg, NULL, 0, 0) == 0:
                    return msg.wParam
                ctypes.windll.user32.TranslateMessage(pMsg)
                ctypes.windll.user32.DispatchMessageA(pMsg)
            else:
                count = timeGetTime()
                d = count - lastCount
                if d > 0:
                    window.controller.onUpdate(d)
                    window.Redraw()
                    lastCount = count

    @staticmethod
    def WndProc(hwnd, message, wParam, lParam):
        # sys.stderr.write("WindowFactory::WndProc\n")
        if message == win32con.WM_CREATE:
            lpcreatestruct = ctypes.cast(lParam, ctypes.POINTER(CREATESTRUCT))
            createstruct = lpcreatestruct.contents
            window = ctypes.cast(createstruct.lpCreateParams, ctypes.py_object).value
            window.hwnd = hwnd
            window.SetLongPtr(hwnd, win32con.GWL_USERDATA, createstruct.lpCreateParams)
            window.SetLongPtr(
                hwnd,
                win32con.GWL_WNDPROC,
                ctypes.cast(WNDPROC(Window.WndProcProxy), ctypes.c_void_p),
            )
            return window.WndProc(hwnd, message, wParam, lParam)

        elif message == win32con.WM_DESTROY:
            ctypes.windll.user32.PostQuitMessage(0)
            return 0

        return DefWindowProc(hwnd, message, wParam, lParam)


class LoopManager:
    def __init__(
        self,
        width=win32con.CW_USEDEFAULT,
        height=win32con.CW_USEDEFAULT,
        wndclass=None,
        title=b"glglue.wgl",
        version=None,
        profile=None,
    ):
        self.factory = WindowFactory()
        self.window: Window = self.factory.create(
            Window, width=width, height=height, wndclass=wndclass, title=title
        )
        self.window.createGLContext(16, version=version, profile=profile)
        ShowWindow(self.window.hwnd, win32con.SW_SHOWNORMAL)
        self.msg = MSG()
        self.pMsg = ctypes.pointer(self.msg)
        self.NULL = ctypes.c_int(win32con.NULL)

        from OpenGL.WGL.EXT.swap_control import wglSwapIntervalEXT

        wglSwapIntervalEXT(1)

    def begin_frame(self) -> Optional[glglue.frame_input.FrameInput]:
        while True:
            if (
                ctypes.windll.user32.PeekMessageA(
                    self.pMsg, self.NULL, 0, 0, win32con.PM_NOREMOVE
                )
                != 0
            ):
                if ctypes.windll.user32.GetMessageA(self.pMsg, self.NULL, 0, 0) == 0:
                    return  # msg.wParam
                ctypes.windll.user32.TranslateMessage(self.pMsg)
                ctypes.windll.user32.DispatchMessageA(self.pMsg)
            break

        frame = glglue.frame_input.FrameInput(
            elapsed_time=datetime.timedelta(milliseconds=timeGetTime()),
            mouse_x=self.window.x,
            mouse_y=self.window.y,
            width=self.window.width,
            height=self.window.height,
            mouse_left=self.window.mouse_left,
            mouse_middle=self.window.mouse_middle,
            mouse_right=self.window.mouse_right,
            mouse_wheel=self.window.wheel,
        )
        self.window.wheel = 0
        return frame

    def end_frame(self):
        ps = PAINTSTRUCT()
        hdc = BeginPaint(self.window.hwnd, ctypes.byref(ps))
        if hdc:
            SwapBuffers(hdc)
        EndPaint(self.window.hwnd, ctypes.byref(ps))


def mainloop(render: glglue.frame_input.RenderFunc, **kw):
    loop = LoopManager(**kw)

    while True:
        frame = loop.begin_frame()
        if not frame:
            break
        render(frame)
        loop.end_frame()



from typing import NamedTuple, Callable, TypeAlias
import datetime


class FrameInput(NamedTuple):
    elapsed_time: datetime.timedelta = datetime.timedelta()
    width: int = 0
    height: int = 0
    mouse_x: int = 0
    mouse_y: int = 0
    mouse_left: bool = False
    mouse_right: bool = False
    mouse_middle: bool = False
    mouse_wheel: int = 0
    is_active: bool = True  # imgui render target
    is_hover: bool = True  # imgui render target


RenderFunc: TypeAlias = Callable[[FrameInput], None]



from typing import List
import ctypes
import glm


class Float3(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_float),
        ('y', ctypes.c_float),
        ('z', ctypes.c_float),
    ]

    def __iter__(self):
        yield self.x
        yield self.y
        yield self.z


class Vertex(ctypes.Structure):
    _fields_ = [
        ('position', Float3),
        ('normal', Float3),
        ('color', Float3),
    ]


class MeshBuilder:
    def __init__(self) -> None:
        self.vertices: List[Vertex] = []

    def push_triangle(self, p0: glm.vec3, p1: glm.vec3, p2: glm.vec3, n: glm.vec3, color: glm.vec3):
        self.vertices.append(
            Vertex(Float3(p0.x, p0.y, p0.z), Float3(n.x, n.y, n.z), Float3(color.x, color.y, color.z)))
        self.vertices.append(
            Vertex(Float3(p1.x, p1.y, p1.z), Float3(n.x, n.y, n.z), Float3(color.x, color.y, color.z)))
        self.vertices.append(
            Vertex(Float3(p2.x, p2.y, p2.z), Float3(n.x, n.y, n.z), Float3(color.x, color.y, color.z)))

    def push_quad(self, p0, p1, p2, p3, color):
        n = glm.cross(glm.normalize(p0-p1), glm.normalize(p2-p1)) # type: ignore
        self.push_triangle(p0, p1, p2, n, color)
        self.push_triangle(p2, p3, p0, n, color)

    def create_vertices(self) -> ctypes.Array:
        vertices = (Vertex * len(self.vertices))(*self.vertices)
        return vertices



from typing import List, Optional
from OpenGL import GL
from glglue.glo.shader import Shader, UniformUpdater
from glglue.glo.vao import Vao


def empty():
    return []


class Submesh:
    def __init__(
        self,
        topology,
        *,
        draw_count=0,
        shader: Optional[Shader] = None,
        props: List[UniformUpdater] = empty()
    ) -> None:
        self.topology = topology
        self.draw_count = draw_count
        assert (not shader) or isinstance(shader, Shader)
        self.shader = shader
        self.properties = props


class Drawable:
    def __init__(self, vao: Vao) -> None:
        self.vao = vao
        self.submeshes: List[Submesh] = []

    def push_submesh(
        self,
        shader: Shader,
        draw_count: int,
        properties: list,
        *,
        topology=GL.GL_TRIANGLES
    ):
        assert isinstance(shader, Shader)
        self.submeshes.append(
            Submesh(topology, shader=shader, draw_count=draw_count, props=properties)
        )

    def draw(self):
        self.vao.bind()
        for submesh in self.submeshes:
            if submesh.shader:
                with submesh.shader:
                    for prop in submesh.properties:
                        prop()
                    self.vao.draw(submesh.draw_count, topology=submesh.topology)
        self.vao.unbind()



# coding: utf-8
'''
pip install pysdl2 pysdl2-dll
'''

def main():
    import glglue.pysdl2
    from glglue.scene.sample import SampleScene

    scene = SampleScene()

    loop = glglue.pysdl2.LoopManager()
    while True:
        frame = loop.begin_frame()
        if not frame:
            break
        scene.render(frame)
        loop.end_frame()


if __name__ == "__main__":
    main()



from typing import Optional, List, Callable, TypeAlias, Protocol
from glglue.frame_input import FrameInput


Callback: TypeAlias = Callable[[FrameInput, Optional[FrameInput]], None]
BeginEndCallback: TypeAlias = Callable[[FrameInput], None]
DragCallback: TypeAlias = Callable[[FrameInput, int, int], None]


class DragInterface(Protocol):
    def begin(self, mouse_input: FrameInput):
        ...

    def drag(self, mouse_input: FrameInput, dx: int, dy: int):
        ...

    def end(self, mouse_input: FrameInput):
        ...


class MouseEvent:
    def __init__(self) -> None:
        self.callbacks: List[Callback] = []
        self.last_input: Optional[FrameInput] = None

        # highlevel event
        self.left_active = None
        self.left_pressed: List[BeginEndCallback] = []
        self.left_drag: List[DragCallback] = []
        self.left_released: List[BeginEndCallback] = []
        self.right_active = None
        self.right_pressed: List[BeginEndCallback] = []
        self.right_drag: List[DragCallback] = []
        self.right_released: List[BeginEndCallback] = []
        self.middle_active = None
        self.middle_pressed: List[BeginEndCallback] = []
        self.middle_drag: List[DragCallback] = []
        self.middle_released: List[BeginEndCallback] = []

        self.wheel: List[Callable[[int], None]] = []
        self.nvg = None

    def __iadd__(self, callback: Callback):
        self.callbacks.append(callback)
        return self

    def process(self, current: FrameInput):
        for callback in self.callbacks:
            callback(current, self.last_input)

        #
        # highlevel event
        #
        dx = 0
        dy = 0
        if self.last_input:
            dx = current.mouse_x - self.last_input.mouse_x
            dy = current.mouse_y - self.last_input.mouse_y
        # pressed
        if current.is_hover:
            if (
                not self.last_input or not self.last_input.mouse_left
            ) and current.mouse_left:
                self.left_active = (current.mouse_x, current.mouse_y)
                for callback in self.left_pressed:
                    callback(current)
            if (
                not self.last_input or not self.last_input.mouse_right
            ) and current.mouse_right:
                self.right_active = (current.mouse_x, current.mouse_y)
                for callback in self.right_pressed:
                    callback(current)
            if (
                not self.last_input or not self.last_input.mouse_middle
            ) and current.mouse_middle:
                self.middle_active = (current.mouse_x, current.mouse_y)
                for callback in self.middle_pressed:
                    callback(current)
        # drag
        if current.is_active:
            if current.mouse_left:
                # self.left_active = True
                for callback in self.left_drag:
                    callback(current, dx, dy)
            if current.mouse_right:
                # self.right_active = True
                for callback in self.right_drag:
                    callback(current, dx, dy)
            if current.mouse_middle:
                # self.middle_active = True
                for callback in self.middle_drag:
                    callback(current, dx, dy)
        # released
        if self.left_active and not current.mouse_left:
            for callback in self.left_released:
                callback(current)
            self.left_active = None
        if self.right_active and not current.mouse_right:
            for callback in self.right_released:
                callback(current)
            self.right_active = None
        if self.middle_active and not current.mouse_middle:
            for callback in self.middle_released:
                callback(current)
            self.middle_active = None

        if current.is_active or current.is_hover:
            if current.mouse_wheel:
                for callback in self.wheel:
                    callback(current.mouse_wheel)

        self.last_input = current

    def bind_left_drag(self, drag_handler: DragInterface):
        self.left_pressed.append(drag_handler.begin)
        self.left_drag.append(drag_handler.drag)
        self.left_released.append(drag_handler.end)

    def bind_right_drag(self, drag_handler: DragInterface):
        self.right_pressed.append(drag_handler.begin)
        self.right_drag.append(drag_handler.drag)
        self.right_released.append(drag_handler.end)

    def bind_middle_drag(self, drag_handler: DragInterface):
        self.middle_pressed.append(drag_handler.begin)
        self.middle_drag.append(drag_handler.drag)
        self.middle_released.append(drag_handler.end)



from .shader import Shader, UniformLocation, UniformBlockIndex, UniformUpdater
from .vbo import Vbo, Ibo
from .vao import Vao
from .vertex_layout import VertexLayout, AttributeLocation
from .texture import Texture
from .fbo import Fbo, FboRenderer

__all__ = [
    "Shader",
    "UniformLocation",
    "UniformBlockIndex",
    "UniformUpdater",
    "Vbo",
    "Ibo",
    "Vao",
    "VertexLayout",
    "AttributeLocation",
    "Texture",
    "Fbo",
    "FboRenderer",
]



import sys
from typing import Optional
import datetime
import os
import pathlib
import platform
import ctypes
from glglue.frame_input import FrameInput

if platform.system() == "Windows":
    if "PYSDL2_DLL_PATH" in os.environ:
        if os.path.exists(os.environ["PYSDL2_DLL_PATH"]):
            # OK
            pass
        else:
            del os.environ["PYSDL2_DLL_PATH"]

    if "PYSDL2_DLL_PATH" not in os.environ:
        try:
            import sdl2dll
        except ImportError:
            exe = pathlib.Path(sys.executable)
            sdl2dll = exe.parent / "sdl2.dll"
            if sdl2dll.exists():
                os.environ["PYSDL2_DLL_PATH"] = str(exe.parent)

from sdl2 import *


class LoopManager:
    def __init__(self, *, resizable=True, **kw):
        title = kw.get("title", b"glglue.pysdl2")

        # window state
        self.width = kw.get("width", 600)
        self.height = kw.get("height", 400)
        self.x = 0
        self.y = 0
        self.left_down = False
        self.right_down = False
        self.middle_down = False

        SDL_Init(SDL_INIT_VIDEO)
        SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1)
        self.window = SDL_CreateWindow(
            title,
            SDL_WINDOWPOS_UNDEFINED,
            SDL_WINDOWPOS_UNDEFINED,
            self.width,
            self.height,
            SDL_WINDOW_OPENGL,
        )
        SDL_GL_CreateContext(self.window)

        if resizable:
            SDL_SetWindowResizable(self.window, True)

        self.event = SDL_Event()

    def begin_frame(self) -> Optional[FrameInput]:
        """
        returns elapsed milliseconds
        """
        wheel = 0
        while SDL_PollEvent(ctypes.byref(self.event)) != 0:
            if self.event.type == SDL_QUIT:
                return
            elif self.event.type == SDL_KEYDOWN:
                # self.controller.onKeyDown(self.event.key.keysym.sym)
                pass
            elif self.event.type == SDL_MOUSEBUTTONDOWN:
                if self.event.button.button == SDL_BUTTON_LEFT:
                    self.left_down = True
                elif self.event.button.button == SDL_BUTTON_MIDDLE:
                    self.middle_down = True
                elif self.event.button.button == SDL_BUTTON_RIGHT:
                    self.right_down = True
            elif self.event.type == SDL_MOUSEBUTTONUP:
                if self.event.button.button == SDL_BUTTON_LEFT:
                    self.left_down = False
                elif self.event.button.button == SDL_BUTTON_MIDDLE:
                    self.middle_down = False
                elif self.event.button.button == SDL_BUTTON_RIGHT:
                    self.right_down = False
            elif self.event.type == SDL_MOUSEMOTION:
                self.x = self.event.motion.x
                self.y = self.event.motion.y
            elif self.event.type == SDL_MOUSEWHEEL:
                wheel = self.event.wheel.y
            elif self.event.type == SDL_WINDOWEVENT:
                if self.event.window.event == SDL_WINDOWEVENT_RESIZED:
                    self.width = self.event.window.data1
                    self.height = self.event.window.data2

        return FrameInput(
            elapsed_time=datetime.timedelta(microseconds=SDL_GetTicks()),
            mouse_x=self.x,
            mouse_y=self.y,
            width=self.width,
            height=self.height,
            mouse_left=self.left_down,
            mouse_middle=self.middle_down,
            mouse_right=self.right_down,
            mouse_wheel=wheel,
        )

    def end_frame(self):
        SDL_GL_SwapWindow(self.window)



from typing import Optional
from OpenGL import GL
import logging
logger = logging.getLogger(__name__)


class Texture:
    def __init__(self, width: int, height: int, data: Optional[bytes] = None, *, pixel_type=GL.GL_RGBA) -> None:
        self.width = width
        self.height = height
        # GL.GL_RGBA(32bit) or GL.GL_RED(8bit graysclale)
        self.pixel_type = pixel_type
        self.handle = GL.glGenTextures(1)
        logger.debug(f'Texture: {self.handle}')

        self.bind()

        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        # GL.glPixelStorei(GL.GL_UNPACK_ROW_LENGTH, width)
        GL.glPixelStorei(GL.GL_UNPACK_SKIP_PIXELS, 0)
        GL.glPixelStorei(GL.GL_UNPACK_SKIP_ROWS, 0)

        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, self.pixel_type,
                        width, height, 0, self.pixel_type, GL.GL_UNSIGNED_BYTE, data)
        GL.glTexParameteri(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        GL.glTexParameteri(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        self.unbind()

    def __del__(self):
        logger.debug(f'{self.handle}')
        GL.glDeleteTextures([self.handle])

    def update(self, x, y, w, h, data):
        self.bind()

        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        GL.glPixelStorei(GL.GL_UNPACK_ROW_LENGTH, self.width)
        GL.glPixelStorei(GL.GL_UNPACK_SKIP_PIXELS, x)
        GL.glPixelStorei(GL.GL_UNPACK_SKIP_ROWS, y)

        GL.glTexSubImage2D(GL.GL_TEXTURE_2D, 0, x, y, w, h,
                           self.pixel_type, GL.GL_UNSIGNED_BYTE, data)

        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 4)
        GL.glPixelStorei(GL.GL_UNPACK_ROW_LENGTH, 0)
        GL.glPixelStorei(GL.GL_UNPACK_SKIP_PIXELS, 0)
        GL.glPixelStorei(GL.GL_UNPACK_SKIP_ROWS, 0)

        self.unbind()

    def bind(self):
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.handle)

    def unbind(self):
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
